#include <stdio.h>
#include <stdlib.h>

/* funzioni per la gestione di un max-heap */

void swap(int *a, int i, int j)
{
	int k = a[i];

	a[i] = a[j];
	a[j] = k;
}

void heapify(int *a, int n, int i)
{
	int l = 2 * i + 1;
	int r = l + 1;
	int m;

	m = (l < n && a[l] > a[i]) ? l : i;
	if(r < n && a[r] > a[m])
		m = r;
	if(m != i) {
		swap(a,i,m);
		heapify(a,n,m);
	}
}

void buildheap(int *a, int n)
{
	int i;

	for(i = n/2-1; i >= 0; i--)
		heapify(a,n,i);
}



/* heapsort */

void heapsort(int *a, int n)
{
	int i;

	buildheap(a,n);
	for(i = n - 1; i; i--) {
		swap(a,0,i);
		heapify(a,i,0);		
	}
}



/* funzioni per la gestione di una coda di max-priorita' */

int maxheap(int *a)
{
	return a[0];
}

int extractmaxheap(int *a, int n) /* n > 0 */
{
	int m = a[0];

	a[0] = a[n-1];
	heapify(a,n-1,0);
	return m;	
} 

void increasevalue(int *a, int i, int k)
{	
	if(k < a[i])
		return; 
	a[i] = k;
	while(i && a[(i-1)/2] < a[i]) {
		swap(a,i,(i-1)/2);
		i = (i-1)/2;
	}
}

int *heapinsert(int *a, int n, int k)
{
	if(!(a = realloc(a, sizeof(int) * ++n))) {
		fprintf(stderr,"Errore di allocazione \n");
		exit(-1);
	}
	a[n-1] = k-1;
	increasevalue(a,n-1,k);
	return a;
}




/* programmino di prova di heapsort e della coda di priorita' */

void stampa(int *a, int n)
{
	int i;

	for(i = 0; i < n; i++)
		printf("[%d]",a[i]);
	putchar('\n');
}

int main(void)
{
	int aux[] = {4,1,3,2,16,9,10,14,8,7};
	int *a = calloc(10, sizeof(int));
	int i;

	
	if(!a) {
		fprintf(stderr,"Errore di allocazione \n");
		exit(-2);
	}
	for(i = 0; i < 10; i++)
		a[i] = aux[i];
	heapsort(a,10);
	stampa(a,10);
	free(a);



	printf("------------------\n");
	
	if(!(a = calloc(10, sizeof(int)))){
		fprintf(stderr,"Errore di allocazione \n");
		exit(-3);
	}
	for(i = 0; i < 10; i++) {
		a = heapinsert(a,i,aux[i]);
		stampa(a,i+1);
		printf("max %d\n",maxheap(a));
	}
	printf("------------------\n");



	/* esempio d'uso di increasevalue: ogni chiave considerata e' reimpostata al proprio quadrato */
	for(i = 0; i < 10; i++) {
		increasevalue(a,i,i*i);
		stampa(a,10);
	}
	printf("------------------\n");

	
	for(i = 0; i < 10; i++)
		printf("{%d}",extractmaxheap(a,10-i));
	putchar('\n');	

	return 0;
}
