#include <stdio.h>
#include <stdlib.h>

/* macro per dim massima nome file dati */
#define MAXSIZE	80

#define min(A,B) ((A) < (B) ? (A) : (B))

/* macro per accedere a un elemento di un array monodimensionale pensato */
/* come tabella bidimensionale con n righe. j,i sono le coordinate riga,colonna dell'elemento */
#define posto(n,j,i) 	((n)*(j)+(i))

/* legge n interi dal file fp e li colloca nell'array (il cui indirizzo iniziale e') tabella */
/* NB: tabella deve essere gia' stato allocato */
void leggitabella(FILE *fp, int *tabella, int n)
{
	int i,j;

	for(j = 0; j < 2; j++)
		for(i = 0; i < n; i++)
			fscanf(fp,"%d", tabella + posto(n,j,i));
}

void stampatabella(int *tabella, int n)
{
	int i,j;

	for(j = 0; j < 2; j++) {
		for(i = 0; i < n; i++)
			printf("%2d,", tabella[posto(n,j,i)]);
		putchar('\n');
	}
}

/* alloca in memoria dinamica tre array */
/* e li carica con i dati letti dal file puntato da fp */
/* il primo e' un array di char che conterra' in seguito la codifica della soluzione */
/* il secondo e' un array di int che conterra' i tempi di lavorazione delle due catene */
/* il terzo e' un array di int che conterra' i tempi di passaggio */
/* Nota che i tre argomenti corrispondenti sono passati per indirizzo, dunque il tipo */
/* dei parametri e' char **s, int **t1, int **t2 */
/* Tale passaggio e' reso necessario perche' calloc assegna dei valori (gli indirizzi iniziali degli array dinamici) */
/* alle variabili puntatore attraverso le quali manipoleremo gli array */
/* dunque il tipo dei parametri deve essere puntatore a puntatore a tipo base */ 
char preparatabelle(FILE *fp, char **s, int **t1, int **t2, int n)
{
	/* controlla che le allocazioni abbiano successo */
	if(!(  
		 (*s = calloc(n, 1)) &&
		 (*t1 = calloc(n * 2, sizeof(int))) &&
             (*t2 = calloc((n-1)* 2, sizeof(int)))
	   ))
		return 0;		
	leggitabella(fp,*t1,n);
	leggitabella(fp,*t2,n-1);
	return 1;
}

/* calcola una catena ottimale e la codifica nell'array sol come sequenza di bit 0,1 */
/* 0 significhera' che l'elemento sara' da prendere dalla prima catena, 1 dalla seconda. */
/* Restituisce il costo totale di tale catena */
int calcolacatena(char *sol, int *t, int *p, int n)
{
	/* allocazione delle tabelle di supporto: */

	/* sum[posto(n,j,i)] conterra' il costo di una catena ottimale che */
      /* parte dall'elemento in posizione (j,i) e arriva alla fine */
	int *sum = calloc(n * 2, sizeof(int));

	/* pred[posto(n,j,i)] conterra' 0 se */
	/* l'elemento di indice i+1 nella catena ottimale calcolata */
	/* stara' sulla stessa catena j */
	/* pred[posto(n-1,j,i)] conterra' 1 se */
      /* l'elemento di indice i+1 nella catena ottimale calcolata */
      /* stara' sull'altra catena rispetto a j (cioe' quella di indice (j+1)%2) */
	char *pred = calloc((n-1) * 2, 1);

	int i,j,k1,k2;

	/* controllo dell'esito delle allocazioni dinamiche */
	if(!sol || !pred || !sum) {
		fprintf(stderr,"Errore di allocazione\n");
		exit(-4);
	}

	/* i due elementi finali (quelli di indice n-1) di sum sono */
      /* copiati dai corrispettivi elementi di t */
	for(j = 0; j < 2;j++)
		sum[posto(n,j,n-1)] = t[posto(n,j,n-1)];
	/* con un ciclo discendente da n-2 a 0, calcolo tutti gli elementi di sum e pred */
	for(i = n - 2; i >= 0; i--) {
		for(j = 0; j < 2; j++) {
			/* k1 e k2 contengono i valori delle due soluzioni possibili per il passo (j,i) */
			/* k1: se scelgo i+1 in modo da proseguire sulla stessa linea j */
			k1 = sum[posto(n,j,i+1)];
			/* k2: se scelgo i+1 in modo da saltare sull'altra linea (j+1)%2 */
			/* in questo caso pago anche il trasferimento p[posto(n-1,j,i)] */
			k2 = sum[posto(n,(j+1)%2,i+1)] + p[posto(n-1,j,i)];
			/* se k2 e' preferita a k1, assegno 1 a pred[posto(n-1,j,i)], altrimenti 0 */
			pred[posto(n-1,j,i)] = (k1 > k2);
			/* sommo a t[posto(n,j,i)] il valore della sottosoluzione migliore */
			sum[posto(n,j,i)] = t[posto(n,j,i)] + min(k1,k2);
		}
	}

	/* fase di ricostruzione della soluzione */
	/* fra le due soluzioni di indice 0 scelgo la migliore */
	/* NB: sum[0] == sum[posto(n,0,0)], sum[n] == sum[posto(n,1,0)] */
	/* sol[0] == 1 sse sum[posto(n,0,0)] e' peggiore di sum[posto(n,1,0)] */
	sol[0] = (sum[0] > sum[n]);
	/* proseguo consultando l'array pred */
	for(i = 1; i < n; i++) {
		/* sol[i] == sol[i-1] sse pred[posto(n-1,sol[i-1],i-1] == 0 */
		sol[i] = (sol[i-1] + pred[posto(n-1,sol[i-1],i-1)]) % 2;
	}

	/* memorizzo in i il costo della soluzione ottimale */
	i = sum[posto(n,sol[0],0)];

	/* libero la memoria utilizzata */
	free(sum);
	free(pred);

	/* restituisco il costo della soluzione ottimale */
	return i;
}

void stampacatena(char *s, int *t, int *p, int n)
{
	int i,j;

	for(j = 0;j < 2;j++) {
		for(i = 0;i < n;i++) {
			if(s[i] == j)
				printf("%2d  ",t[posto(n,j,i)]);
			else
				printf("--  ");
		}
		putchar('\n');
		/* stampo la linea intermedia che visualizza i "salti di catena" */
		if(!j) {
			printf("  ");
			for(i = 0;i < n-1;i++)
				if(s[i] != s[i+1])
					printf("%2d  ",p[posto(n-1,s[i],i)]);
				else
					printf("    ");
			putchar('\n');
		}
		
	}
}

int main(void)
{
	FILE *fr;
	int n, tot;
	int *tempi_lavorazione;
	int *tempi_passaggio;
	char *catena;
	char nomefile[MAXSIZE];

	printf("Immetti nome file dati:");
	scanf("%s",nomefile);

	/* controllo sull'accessibilita' del file da leggere */
	if(!(fr  = fopen(nomefile,"r"))) {
		fprintf(stderr,"file %s non trovato\n",nomefile);
		exit(-2);
	}
	
	/* il file da leggere si assume essere corretto in formato */
      /* se non facessimo questa assunzione dovremmo controllare l'esito di ogni singola fscanf */
	fscanf(fr,"%d",&n);
	/* nota che a preparatabelle si passano gli indirizzi di catena, tempi_lavorazione, tempi_passaggio */
      /* questo perche' preparatabelle dovra' assegnare alle variabili  catena, tempi_lavorazione, tempi_passaggio */
      /* gli indirizzi iniziali degli array che preparera' in memoria dinamica tramite calloc */
      /* Il tipo di &catena e' dunque char **, quello di &tempi_lavorazione e di &tempi_passaggio e' int **. */
	if(!preparatabelle(fr,&catena,&tempi_lavorazione,&tempi_passaggio,n)) {
		fprintf(stderr,"Errore di allocazione\n");
		exit(-3);
	}
	
	tot = calcolacatena(catena,tempi_lavorazione,tempi_passaggio,n);
	printf("\nCosto totale: %d\n\n",tot);
	stampacatena(catena,tempi_lavorazione,tempi_passaggio,n);	

	/* le seguenti 4 istruzioni non sono strettamente necessarie */
	/* le risorse allocate vengono comunque liberate al termine del processo */
	free(catena);
	free(tempi_lavorazione);
	free(tempi_passaggio);
	fclose(fr);

	return 0;
}
