#include <stdio.h>
#include <stdlib.h>

/* legge la scacchiera dal file fp, la alloca e ne restituisce l'indirizzo base */
short *leggiscacchiera(FILE *fp, short r, short c)
{
	short i,j,*scacchiera;
	
	if(!(scacchiera = calloc(r * c, sizeof(short)))) {
		fprintf(stderr,"Errore nell'allocazione della memoria dinamica\n");
		exit(-1);
	}
	for(i = 0; i < r; i++)
		for(j = 0; j < c; j++)
			fscanf(fp,"%hd", scacchiera + (i * c + j));
	return scacchiera;
}

/* calcola la tabella contenente i valori dei percorsi ottimali e ne restituisce l'indirizzo base */
short *calcolapercorso(short *scacchiera, short r, short c)
{
	short i,j,p1,p2,p,*percorso;
	
	if(!(percorso = calloc(r * c, sizeof(short)))) {
		fprintf(stderr,"Errore nell'allocazione della memoria dinamica\n");
		exit(-2);
	}
	
	/* ciclo sulle righe e sulle colonne */
	for(i = 0; i < r; i++)
		for(j = 0; j < c; j++) {
			
			/* determino il valore p del percorso ottimale fino alla casella precedente (i,j), da aggiungere al valore di scacchiera[i][j] */
			
			/* caso 0,0: la casella iniziale */
			if(!i && !j)
				p = 0;
				
			/* caso 0,j: il bordo alto della scacchiera */
			else if(!i)
				p = percorso[j-1];
				
			/* caso i,0: il bordo sinistro della scacchiera */
			else if(!j)
			    p = percorso[(i-1) * c];
				
			/* caso generico: nel bel mezzo della scacchiera */
			else {
				/* determino il migliore dei due percorsi che arrivano in (i,j) */
				p1 = percorso[(i-1) * c + j];
				p2 = percorso[i * c + (j-1)];
				p = (p1 < p2 ? p1 : p2);
			}
			/* memorizzo il costo del percorso di costo ottimo che arriva in (i,j) */
			percorso[i * c + j] = scacchiera[i * c + j] + p;
		}
	return percorso;
}

/* ricostruisce il percorso di costo ottimo, lo stampa e ne restituisce il costo */
int stampapercorso(short *scacchiera, short *percorso, short r, short c)
{
	int value = *scacchiera;
	short i,j,k,l,v;
	char *path = calloc(r + c - 2, sizeof(char));
	
	if(!path) {
		fprintf(stderr,"Errore nell'allocazione della memoria dinamica\n");
		exit(-2);
	}
	
	/* ciclo con k sulla lunghezza del percorso */
	/* iniziando dall'ultimo elemento della sequenza, cioe' (r-1,c-1). */
	/* Determino a ogni passo la posizione dell'elemento precedente della sequenza */
	/* e la memorizzo nell'array path: 1 significa che la precedente sta sopra, 0: che sta a sinistra */
	for(k = r + c - 3, i = r - 1, j = c - 1; k >= 0; k--) {
	
		/* caso bordo alto, non posso che andare a sinistra */
		if(!i) {
			path[k] = 0;
			j--;
			
		/* caso bordo sinistro, non posso che andare in alto */
		} else if(!j) {
			path[k] = 1;
			i--;
			
		/* caso generico: determino il precedente sottraendo il valore del percorso in i-1,j a quello in i,j */
		} else {
			if(percorso[i * c + j] - percorso[(i-1) * c + j] == scacchiera[(i-1) * c + j]) {
				path[k] = 1;
				i--;
			} else {
				path[k] = 0;
				j--;
			}
		}
	}
	
	/* per stampare il percorso */
	/* scandisco l'array path da sinistra verso destra */
	printf("[%2d]",*scacchiera);
	for(k = i = j = 0; k < r + c - 2; k++) {
		
		/* caso prossima casella giace sotto */ 
		if(path[k]) {
			value += v = scacchiera[++i * c + j];
			putchar('\n');
			/* stampo gli spazi necessari per allineare bene l'output */
			for(l = 0; l < j; l++) 
				printf("    ");
				
		/* caso prossima casella giace a destra */
		} else
			value += v = scacchiera[i * c + ++j];
		printf("[%2d]",v);
	}
	free(path);
	return value;
}

int main(int ac, char **av)
{
	FILE *fp;
	short r,c,*scacchiera,*percorso;
	int value;
	
	if(ac != 2) {
		fprintf(stderr,"Uso: %s <nomefile>\n",av[0]);
		return -10;
	}
	if(!(fp = fopen(av[1],"r"))) {
		fprintf(stderr,"File %s non trovato\n",av[1]);
		return -11;
	}
	
	fscanf(fp,"%hd%hd",&r,&c);
	scacchiera = leggiscacchiera(fp,r,c);
	percorso = calcolapercorso(scacchiera,r,c);
	putchar('\n');
	value = stampapercorso(scacchiera,percorso,r,c);
	printf("\n\nCosto del percorso: %d\n",value);
	
	free(percorso);
	free(scacchiera);
	fclose(fp);
	return 0;
}
